package cloud;
import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.lang.*;
import java.text.DecimalFormat;
import cloud.atom.Atom;
import cloud.atom.AtomInfo;
import gui.BasicViewCanvas;

//--------------------------------------------------------------------//
/**
 * dq_\LoX
 */
public class CloudViewCanvas extends BasicViewCanvas{

/** Bohr a (nm) */
  public static final double A0=0.052918;

/** zvZf[^_ */
  public static final int N=320;

/** <i>x</i> W̍ŏl */
  public static final double X_MIN=-10.0;

/** <i>x</i> W̍ől */
  public static final double X_MAX=6.0;

/** <i>x</i> W̍ݕ */
  public static final double DX=(X_MAX-X_MIN)/(double)N;

/** vZ̃bZ[W */
  public static final String CALC_MESSAGE="Now calculating...";

/** l */
  protected static final DecimalFormat
    FORM=new DecimalFormat("#0.0#####");

/** ʃTCY */
  protected int size;

/** ʒS_W */
  protected Point center;

/** tHggbNX */
  protected FontMetrics fm;

/** wiF */
  protected Color bgColor=Color.BLACK;

/** OiF */
  protected Color fgColor=Color.WHITE;

/** q */
  protected AtomInfo info;

/** \錴q */
  protected Atom atom;

/** z֐ */
  protected DistributionFunction dfunc;

/** G[bZ[W */
  protected String message;

/**
 * dq_\LoX\z
 * @param info q
 * @param message G[bZ[W
 */
  public CloudViewCanvas(AtomInfo info,String message){
    this.message=message;
    setAtom(info);
  }

//--------------------------------------------------------------------//
/**
 * `悷錴qݒ肷
 * @param info q
 */
  public void setAtom(AtomInfo info){
    Graphics g=getGraphics();
    if(g!=null){
      g.setColor(fgColor);
      g.drawString(CALC_MESSAGE,3,20);
    }
    this.info=info;
    if(info!=null){
      atom=info.createAtom(N,X_MIN,DX);
      atom.calculate();
      dfunc=atom.getDistributionFunctionX();
    }
  }

//--------------------------------------------------------------------//
/**
 * ĕ`悷
 * @param g `悷OtBbNXReLXg
 */
  public void update(Graphics g){
    if(info!=null){
      paintCloud(bg);
    }
    else if(message!=null){
      bg.setColor(fgColor);
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * `悷
 * @param g `悷OtBbNXReLXg
 */
  public void paint(Graphics g){
    setPaint();
    if(info!=null){
      paintCloud(bg);
    }
    else if(message!=null){
      bg.setColor(fgColor);
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * ʂ̐ݒ
 */
  protected void setPaint(){
    super.setPaint();
    size=Math.min(panelSize.width,panelSize.height);
    center=new Point(panelSize.width/2,panelSize.height/2);
    fm=bg.getFontMetrics();
  }

//--------------------------------------------------------------------//
/**
 * dq_`悷
 * @param g `悷OtBbNXReLXg
 */
  protected void paintCloud(Graphics g){
    g.setColor(bgColor);
    g.fillRect(0,0,panelSize.width,panelSize.height);
    double coef=2.0*info.getRmax()/(double)size;
    double dist[][]=new double[panelSize.width][panelSize.height];
    double distMax=-100.0;
    for(int i=0;i<panelSize.width;i++){
      double x=coef*Math.abs((double)(i-center.x));
      for(int j=0;j<panelSize.height;j++){
      	double y=coef*(double)(j-center.y);
      	double r=Math.sqrt(x*x+y*y);
      	double xd=info.rToX(r);
      	double phy=Math.atan2(y,x);
      	dist[i][j]=Math.log(Math.max(dfunc.get(xd,0.0,phy),1.0E-20));
      	distMax=Math.max(distMax,dist[i][j]);
      }
    }
    coef=info.getViewCoef();
    for(int i=0;i<panelSize.width;i++){
      for(int j=0;j<panelSize.height;j++){
      	float cc=(float)Math.max(1.0-coef*(distMax-dist[i][j]),0.0);
      	g.setColor(new Color(cc,cc,cc));
      	g.drawLine(i,j,i,j);
      }
    }
    drawScale(g);
  }

//--------------------------------------------------------------------//
/**
 * ڐ\
 * @param g `悷OtBbNXReLXg
 */
  protected void drawScale(Graphics g){
    g.setColor(fgColor);
    double maxSize=0.6*(double)size;
    double scale1=(double)size/(2.0*info.getRmax()*A0);
    double scale=1.0;
    while(scale1>maxSize){
      scale1*=0.1;
      scale*=0.1;
    }
    String scales=FORM.format(scale)+"nm";
    int scaleh=(int)(0.5*scale1);
    int sheight=fm.getHeight();
    int height=panelSize.height-2*sheight;
    int width=fm.stringWidth(scales);
    g.fillRect(center.x-scaleh,height,(int)scale1,2);
    g.drawLine(center.x-scaleh,height-2,center.x-scaleh,height+4);
    g.drawLine(center.x+scaleh,height-2,center.x+scaleh,height+4);
    g.drawString(scales,center.x-width/2,height+sheight);
  }
}
