package cloud.atom;
import java.lang.*;
import java.util.Arrays;
import cloud.DistributionFunction;
import cloud.distribution.SingleDistributionFunction;
import math.AssociatedLegendre;
import math.ConstantFunction;
import math.CubicSplineIP;
import math.Function;

//--------------------------------------------------------------------//
/**
 * dqONX
 */
public class Orbit{

/** A_X|oVtH[X|[g@̗\qWi[z */
  public static final double CP[][]=
    {{1.0},
     {1.5,-0.5},
     {23.0/12.0,-16.0/12.0,5.0/12.0},
     {55.0/24.0,-59.0/24.0,37.0/24.0,-9.0/24.0}};

/** A_X|oVtH[X|[g@̕␳qWi[z */
  public static final double CC[][]=
    {{0.5,0.5},
     {0.5,0.5},
     {5.0/12.0,8.0/12.0,-1.0/12.0},
     {9.0/24.0,19.0/24.0,-5.0/24.0,1.0/24.0},
     {251.0/720.0,646.0/720.0,-264.0/720.0,106.0/720.0,-19.0/720.0}};

/** oxg֐̌W */
  public static final double CTHETA=1.0/Math.sqrt(2.0*Math.PI);

/** GlM[ */
  public double ee;

/** qԍ */
  protected int z;

/** f[^̐ */
  protected int nn;

/** dq */
  public ElectricState state;

/** g֐i[z */
  public double radius[];

/** g֐ <i>x</i> i[z */
  protected double draddx[];

/** VfBK[̌W */
  protected double l2l;

/** kdx */
  public double ecount;

/** 臒l */
  protected double tol;

/**
 * dqO\z
 * @param z qԍ
 * @param nn f[^̐
 * @param state dq
 */
  public Orbit(int z,int nn,ElectricState state){
    this.z=z;
    this.nn=nn;
    this.state=state;
    ee=-(double)(z*z)/(double)(2*state.n*state.n);
    l2l=(double)(state.l*(state.l+1));
    radius=new double[nn+1];
    draddx=new double[nn+1];
    ecount=(double)state.ecount;
    tol=1.0E-8;
  }

//--------------------------------------------------------------------//
/**
 * g֐vZ
 * @param r ai[z
 * @param u qj̃N[|eVi[z
 * @param dx <i>x</i> ̍ݕ
 */
  public void calcWaveFuction(double r[],double u[],double dx){
    int trueNode=state.n-state.l-1;
    double dee=0.0;
    radius[0]=dx;
    draddx[0]=dx;
    radius[nn]=dx;
    draddx[nn]=dx;
    do{
      int fit=search(r,u);
      int node=abm1(fit,r,u,dx);
      double ga=radius[fit];
      double loga=draddx[fit]/(r[fit]*radius[fit]);
      node+=abm2(fit,r,u,dx);
      double gb=radius[fit];
      double logb=draddx[fit]/(r[fit]*radius[fit]);
      modify(fit,ga/gb);
      normalize(r,dx);
      if(node!=trueNode){
      	dee=(double)(trueNode-node)/(double)(state.n*state.n*state.n);
      }
      else{
      	double detemp=-0.5*radius[fit]*radius[fit]*(logb-loga);
      	dee=Math.min(Math.abs(detemp),1.0E-2);
      	if(detemp<0.0) dee=-dee;
      }
      ee+=dee;
      if(ee>0.0){
      	System.out.println("Energy can not converge.");
      	System.exit(1);
      }
    }while(Math.abs(dee)>tol);
  }

//--------------------------------------------------------------------//
/**
 * K_T
 * @param r ai[z
 * @param u qj̃N[|eVi[z
 * @return K_̃CfbNX
 */
  protected int search(double r[],double u[]){
    int index=nn;
    while((index>0) && (ke(index,r,u)>0.0)){
      index--;
    }
    return index;
  }

//--------------------------------------------------------------------//
/**
 * A_X|oVtH[X|[g@Ō_ߓ_܂łvZ
 * @param fit K_̃CfbNX
 * @param r ai[z
 * @param u qj̃N[|eVi[z
 * @param dx <i>x</i> ̍ݕ
 */
  protected int abm1(int fit,double r[],double u[],double dx){
    double draddx2[]=new double[nn+1];
    draddx2[0]=dif2(0,r,u);
    int node=0;
    for(int i=1;i<=fit;i++){
      double drp=0.0;
      double dr2p=0.0;
      int i2=Math.min(i-1,3);
      for(int j=0;j<CP[i2].length;j++){
      	drp+=CP[i2][j]*draddx[i-j-1];
      	dr2p+=CP[i2][j]*draddx2[i-j-1];
      }
      radius[i]=radius[i-1]+dx*drp;
      draddx[i]=draddx[i-1]+dx*dr2p;
      draddx2[i]=dif2(i,r,u);
      drp=0.0;
      dr2p=0.0;
      i2=Math.min(i-1,4);
      for(int j=0;j<CC[i2].length;j++){
      	drp+=CC[i2][j]*draddx[i-j];
      	dr2p+=CC[i2][j]*draddx2[i-j];
      }
      radius[i]=radius[i-1]+dx*drp;
      draddx[i]=draddx[i-1]+dx*dr2p;
      if(radius[i]*radius[i-1]<0.0) node++;
    }
    return node;
  }

//--------------------------------------------------------------------//
/**
 * A_X|oVtH[X|[g@Ŗ_ߓ_܂łvZ
 * @param fit K_̃CfbNX
 * @param r ai[z
 * @param u qj̃N[|eVi[z
 * @param dx <i>x</i> ̍ݕ
 */
  protected int abm2(int fit,double r[],double u[],double dx){
    double draddx2[]=new double[nn+1];
    draddx2[nn]=dif2(nn,r,u);
    int node=0;
    for(int i=nn-1;i>=fit;i--){
      double drp=0.0;
      double dr2p=0.0;
      int i2=Math.min(nn-1-i,3);
      for(int j=0;j<CP[i2].length;j++){
      	drp-=CP[i2][j]*draddx[i+j+1];
      	dr2p-=CP[i2][j]*draddx2[i+j+1];
      }
      radius[i]=radius[i+1]+dx*drp;
      draddx[i]=draddx[i+1]+dx*dr2p;
      draddx2[i]=dif2(i,r,u);
      drp=0.0;
      dr2p=0.0;
      i2=Math.min(nn-1-i,4);
      for(int j=0;j<CC[i2].length;j++){
      	drp-=CC[i2][j]*draddx[i+j];
      	dr2p-=CC[i2][j]*draddx2[i+j];
      }
      radius[i]=radius[i+1]+dx*drp;
      draddx[i]=draddx[i+1]+dx*dr2p;
      if(radius[i]*radius[i+1]<0.0) node++;
    }
    return node;
  }

//--------------------------------------------------------------------//
/**
 * 2 K̒lԂ
 * @param index f[^̃CfbNX
 * @param r ai[z
 * @param u qj̃N[|eVi[z
 * @return 2K̒l
 */
  protected double dif2(int index,double r[],double u[]){
    return ke(index,r,u)*radius[index]+draddx[index];
  }

//--------------------------------------------------------------------//
/**
 * 2 K̒̉Zq̒lԂ
 * @param index f[^̃CfbNX
 * @param r ai[z
 * @param u qj̃N[|eVi[z
 * @return 2K̒̉Zq̒l
 */
  protected double ke(int index,double r[],double u[]){
    return 2.0*r[index]*(u[index]-r[index]*ee)+l2l;
  }

//--------------------------------------------------------------------//
/**
 * g֐C
 * @param fit K_̃CfbNX
 * @param coef ␳W
 */
  protected void modify(int fit,double coef){
    for(int i=fit;i<=nn;i++){
      radius[i]*=coef;
      draddx[i]*=coef;
    }
  }

//--------------------------------------------------------------------//
/**
 * g֐Ki
 * @param r ai[z
 * @param dx <i>x</i> ̍ݕ
 */
  protected void normalize(double r[],double dx){
    double ss=0.0;
    for(int i=0;i<=nn;i++){
      ss+=radius[i]*radius[i]*r[i];
    }
    ss=1.0/Math.sqrt(ss*dx);
    for(int i=0;i<=nn;i++){
      radius[i]*=ss;
      draddx[i]*=ss;
    }
  }

//--------------------------------------------------------------------//
/**
 * GlM[XV
 * @param du qj̃N[|eV̕ωʂi[z
 * @param dx <i>x</i> ̍ݕ
 */
  public void updateEnergy(double du[],double dx){
    double dee=0.0;
    for(int i=0;i<=nn;i++){
      dee+=radius[i]*radius[i]*du[i];
    }
    ee+=dee*dx;
  }

//--------------------------------------------------------------------//
/**
 * dq 1 ̓dqxԂ
 * @param index dqx̃CfbNX
 * @return w肳ꂽʒu̓dq 1 ̓dqx
 */
  public double singleDensity(int index){
    return radius[index]*radius[index];
  }

//--------------------------------------------------------------------//
/**
 * dqxԂ
 * @param index dqx̃CfbNX
 * @return w肳ꂽʒu̓dqx
 */
  public double density(int index){
    return ecount*radius[index]*radius[index];
  }

//--------------------------------------------------------------------//
/**
 * GlM[Ԃ
 * @return dqÕGlM[
 */
  public double energy(){
    return ecount*ee;
  }

//--------------------------------------------------------------------//
/**
 * ag֐Ԃ
 * @param r ai[z
 * @return ag֐
 */
  public Function getWaveFunction(double r[]){
    double f[]=new double[nn+1];
    for(int i=0;i<=nn;i++){
      f[i]=radius[i]/r[i];
    }
    return new CubicSplineIP(nn+1,r,f);
  }

//--------------------------------------------------------------------//
/**
 * ag֐Ԃ
 * @param x <i>x</i> i[z
 * @return ag֐
 */
  public Function getWaveFunctionX(double x[]){
    double f[]=new double[nn+1];
    for(int i=0;i<=nn;i++){
      f[i]=Math.sqrt((double)z)*radius[i]*Math.exp(-x[i]);
    }
    return new CubicSplineIP(nn+1,x,f);
  }

//--------------------------------------------------------------------//
/**
 * oxg֐Ԃ
 * @return oxg֐
 */
  public Function getThetaFunction(){
    return new ConstantFunction(CTHETA);
  }

//--------------------------------------------------------------------//
/**
 * ʊpg֐Ԃ
 * @param count f[^
 * @return ʊpg֐i[z
 */
  public Function[] getPhyFunction(int count){
    Function func[]=new Function[state.ecount];
    double dphy=Math.PI/(double)count;
    double phy[]=new double[count+1];
    for(int j=0;j<=count;j++){
      phy[j]=dphy*(double)j-0.5*Math.PI;
    }
    for(int i=0;i<state.ecount;i++){
      int m=Math.abs(ElectricState.M[i]);
      double clm=1.0;
      for(int j=state.l-m;j<state.l+m;j++) clm*=(double)(j+1);
      double cth=Math.sqrt(((double)state.l+0.5)/clm);
      double f[]=new double[count+1];
      for(int j=0;j<=count;j++){
      	f[j]=cth*AssociatedLegendre.get(Math.sin(phy[j]),state.l,m);
      }
      func[i]=new CubicSplineIP(count+1,phy,f);
    }
    return func;
  }

//--------------------------------------------------------------------//
/**
 * ag֐2Ԃ
 * @param r ai[z
 * @return ag֐2
 */
  public Function getWaveFunction2(double r[]){
    double f[]=new double[nn+1];
    for(int i=0;i<=nn;i++){
      double f1=radius[i]/r[i];
      f[i]=f1*f1;
    }
    return new CubicSplineIP(nn+1,r,f);
  }

//--------------------------------------------------------------------//
/**
 * ag֐2Ԃ
 * @param x <i>x</i> i[z
 * @return ag֐2
 */
  public Function getWaveFunctionX2(double x[]){
    double f[]=new double[nn+1];
    for(int i=0;i<=nn;i++){
      f[i]=(double)z*radius[i]*radius[i]*Math.exp(-2.0*x[i]);
    }
    return new CubicSplineIP(nn+1,x,f);
  }

//--------------------------------------------------------------------//
/**
 * oxg֐2Ԃ
 * @return oxg֐2
 */
  public Function getThetaFunction2(){
    return new ConstantFunction(CTHETA*CTHETA);
  }

//--------------------------------------------------------------------//
/**
 * ʊpg֐2Ԃ
 * @param count f[^
 * @return ʊpg֐2i[z
 */
  public Function[] getPhyFunction2(int count){
    Function func[]=new Function[state.ecount];
    double dphy=Math.PI/(double)count;
    double phy[]=new double[count+1];
    for(int j=0;j<=count;j++){
      phy[j]=dphy*(double)j-0.5*Math.PI;
    }
    for(int i=0;i<state.ecount;i++){
      int m=Math.abs(ElectricState.M[i]);
      double clm=1.0;
      for(int j=state.l-m;j<state.l+m;j++) clm*=(double)(j+1);
      double cth=Math.sqrt(((double)state.l+0.5)/clm);
      double f[]=new double[count+1];
      for(int j=0;j<=count;j++){
      	double f1=cth*AssociatedLegendre.get(Math.sin(phy[j]),
      	      	      	      	      	     state.l,m);
      	f[j]=f1*f1;
      }
      func[i]=new CubicSplineIP(count+1,phy,f);
    }
    return func;
  }

//--------------------------------------------------------------------//
/**
 * ʊpg֐2aԂ
 * @param count f[^
 * @return ʊpg֐2a
 */
  public Function getPhyFunction2sum(int count){
    double dphy=Math.PI/(double)count;
    double phy[]=new double[count+1];
    for(int j=0;j<=count;j++){
      phy[j]=dphy*(double)j-0.5*Math.PI;
    }
    double f[]=new double[count+1];
    Arrays.fill(f,0.0);
    for(int i=0;i<state.ecount;i++){
      int m=Math.abs(ElectricState.M[i]);
      double clm=1.0;
      for(int j=state.l-m;j<state.l+m;j++) clm*=(double)(j+1);
      double cth=Math.sqrt(((double)state.l+0.5)/clm);
      for(int j=0;j<=count;j++){
      	double f1=cth*AssociatedLegendre.get(Math.sin(phy[j]),
      	      	      	      	      	     state.l,m);
      	f[j]+=f1*f1;
      }
    }
    return new CubicSplineIP(count+1,phy,f);
  }

//--------------------------------------------------------------------//
/**
 * z֐Ԃ
 * @param r ai[z
 * @param count ʊpf[^
 * @return z֐
 */
  public DistributionFunction getDistributionFunction(double r[],
      	      	      	      	      	      	      int count){
    return new SingleDistributionFunction(getWaveFunction2(r),
      	      	      	      	      	  getThetaFunction2(),
      	      	      	      	      	  getPhyFunction2sum(count));
  }

//--------------------------------------------------------------------//
/**
 * z֐Ԃ
 * @param x <i>x</i> i[z
 * @param count ʊpf[^
 * @return z֐
 */
  public DistributionFunction getDistributionFunctionX(double x[],
      	      	      	      	      	      	       int count){
    return new SingleDistributionFunction(getWaveFunctionX2(x),
      	      	      	      	      	  getThetaFunction2(),
      	      	      	      	      	  getPhyFunction2sum(count));
  }
}
